# coding=utf-8
# Copyright 2019 YAM AI Machinery Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from mlt.formatter import MultiLabelText, FromFastText, Normalizer, Formatter, ToFastText
from common.ex import YamconvError
import sqlite3
import logging
import os


class Converter:
    def __init__(self, reader, from_formatter, writer, to_formatter,
                 logger=None, nlines=1000):
        self.reader = reader
        self.from_formatter = from_formatter
        self.writer = writer
        self.to_formatter = to_formatter
        self.logger = logger
        self.nlines = nlines

    def info(self, msg):
        if self.logger:
            self.logger.info(msg)

    def err(self, msg):
        if self.logger:
            self.logger.error(msg)
        raise YamconvError(msg)

    def convert(self):
        try:
            self.reader.open()
        except Exception as e:
            self.err('Error opening input file {}: {}'.format(
                self.reader.filepath, e))
        self.info('Opened input file {}.'.format(self.reader.filepath))
        try:
            self.writer.open()
        except Exception as e:
            self.err('Error opening output file {}: {}'.format(
                self.writer.filepath, e))
        self.info('Opened output file {}.'.format(self.writer.filepath))
        i = 0
        while True:
            try:
                from_mlt = self.reader.read()
            except Exception as e:
                self.err('Error reading input file {}: {}'.format(
                    self.reader.filepath, e))
            if not from_mlt:
                break
            norm_mlt = self.from_formatter.format(from_mlt)
            to_mlt = self.to_formatter.format(norm_mlt)
            try:
                self.writer.write(to_mlt)
            except Exception as e:
                self.err('Error writing output file {}: {}'.format(
                    self.writer.filepath, e))
            i += 1
            if i % 1000 == 0:
                self.info('Processed {} records.'.format(i))
        self.info('Completed processing {} records in total.'.format(i))
        try:
            self.reader.close()
        except Exception as e:
            self.err('Error closing input file {}: {}'.format(
                self.reader.filepath, e))
        self.info('Closed input file {}.'.format(self.reader.filepath))
        try:
            self.writer.close()
        except Exception as e:
            self.err('Error closing output file {}: {}'.format(
                self.writer.filepath, e))
        self.info('Closed output file {}.'.format(self.writer.filepath))


class Reader(ABC):
    def __init__(self, filepath):
        self.filepath = filepath

    def open():
        pass

    def read():
        pass

    def close():
        pass


class Writer(ABC):
    def __init__(self, filepath):
        self.filepath = filepath

    def open():
        pass

    def write(mlt):
        pass

    def close():
        pass


class FastTextReader(Reader):
    def __init__(self, fasttext_path):
        super(self.__class__, self).__init__(fasttext_path)

    def open(self):
        if not os.path.isfile(self.filepath):
            raise YamconvError(
                'Input file {} does not exists.'.format(self.filepath))
        self.fasttext_file = open(self.filepath, 'r')

    def read(self):
        line = self.fasttext_file.readline()
        if line == '':
            return None
        tokens = line.split()
        words = set()
        mlt = MultiLabelText()
        for token in tokens:
            is_text = False
            if not is_text:
                if token.startswith('__label__'):
                    mlt.add_label(token)
                else:
                    is_text = True
                    mlt.add_word(token)
            else:
                mlt.add_word(token)
        return mlt

    def close(self):
        self.fasttext_file.close()


class FastTextWriter(Writer):
    def __init__(self, fasttext_path):
        super(self.__class__, self).__init__(fasttext_path)

    def open(self):
        self.fasttext_file = open(self.filepath, 'w')

    def write(self, mlt):
        print(' '.join(list(mlt.labels) + [mlt.text]), file=self.fasttext_file)

    def close(self):
        self.fasttext_file.close()


class SQLiteReader(Reader):
    def __init__(self, sqlite_path):
        super(self.__class__, self).__init__(sqlite_path)

    def open(self):
        if not os.path.isfile(self.filepath):
            raise YamconvError(
                'Input file {} does not exists.'.format(self.filepath))
        self.conn = sqlite3.connect(self.filepath)
        self.cur = self.conn.cursor()
        self.cur.execute('SELECT id FROM texts')
        rows = self.cur.fetchall()
        self.text_ids = [row[0] for row in rows]

    def read(self):
        try:
            text_id = self.text_ids.pop()
        except:
            return None
        self.cur.execute('SELECT text FROM texts WHERE id = ?', (text_id, ))
        mlt = MultiLabelText(self.cur.fetchone()[0])
        self.cur.execute(
            'SELECT label FROM labels WHERE text_id = ?', (text_id, ))
        rows = self.cur.fetchall()
        for row in rows:
            mlt.add_label(row[0])
        return mlt

    def close(self):
        self.conn.close()


schema = '''
        DROP TABLE IF EXISTS texts;
        CREATE TABLE texts (
            id TEXT NOT NULL PRIMARY KEY,
            text TEXT NOT NULL
        );
        DROP TABLE IF EXISTS labels;
        CREATE TABLE labels (
            label TEXT NOT NULL,
            text_id text NOT NULL,
            FOREIGN KEY (text_id) REFERENCES texts(id)
        );
        DROP INDEX IF EXISTS label_index;
        CREATE INDEX label_index ON labels (label);
        CREATE INDEX text_id_index ON labels (text_id);
    '''


class SQLiteWriter(Writer):
    def __init__(self, sqlite_path):
        super(self.__class__, self).__init__(sqlite_path)

    def open(self):
        self.conn = sqlite3.connect(self.filepath)
        # Can use autocommit for faster performance
        # self.conn.isolation_level = None
        self.cur = self.conn.cursor()
        self.cur.executescript(schema)
        self.text_id = 0

    def write(self, mlt):
        self.cur.execute(
            'INSERT INTO texts (id, text) VALUES (?, ?)', (self.text_id, mlt.text))
        for label in mlt.labels:
            self.cur.execute(
                'INSERT INTO labels (label, text_id) VALUES (?, ?)', (self.text_id, label))
        self.conn.commit()  # Can omit this commit for autocommit
        self.text_id += 1

    def close(self):
        self.conn.commit()
        self.conn.close()


class FastText2SQLite(Converter):
    def __init__(self, fasttext_path, sqlite_path, cache_labels=True,
                 logger=None, nlines=1000):
        reader = FastTextReader(fasttext_path)
        from_formatter = FromFastText(cache_labels=cache_labels)
        writer = SQLiteWriter(sqlite_path)
        to_formatter = Normalizer(cache_labels=cache_labels)
        super(self.__class__, self).__init__(
            reader, from_formatter, writer, to_formatter, logger, nlines)


class SQLite2FastText(Converter):
    def __init__(self, sqlite_path, fasttext_path, cache_labels=True,
                 logger=None, nlines=1000):
        reader = SQLiteReader(sqlite_path)
        from_formatter = Formatter(cache_labels=cache_labels)
        writer = FastTextWriter(fasttext_path)
        to_formatter = ToFastText(cache_labels=cache_labels)
        super(self.__class__, self).__init__(
            reader, from_formatter, writer, to_formatter, logger, nlines)



# coding=utf-8
# Copyright 2019 YAM AI Machinery Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains functions to clean and tokenize text.
>>> s = u'''世界，你好！
... Hello, World!'''
>>> t = remove_symbols(s)
>>> t == u'世界 你好  Hello  World '
True
>>> u = tokenize_unicode_chars(t)
>>> u == '世 界   你 好     Hello    World  '
True
>>> v = compact_spaces(u)
>>> v == u'世 界 你 好 Hello World'
True
>>> normalize(s)
'世 界 你 好 hello world'
"""

import re


def remove_symbols(s):
    return re.sub(r'[^\w]', ' ', s, flags=re.UNICODE)


def tokenize_unicode_chars(s):
    return re.sub(r'([^A-Za-z])', '\\1 ', s, flags=re.UNICODE)


def compact_spaces(s):
    return ' '.join(s.split())


def normalize(s):
    return compact_spaces(tokenize_unicode_chars(remove_symbols(s))).lower()


def normalize_label(s):
    return '_'.join(compact_spaces(remove_symbols(s)).lower().split())



# coding=utf-8
# Copyright 2019 YAM AI Machinery Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
>>> text = 'this is a text'
>>> labels = { 'Hello,  World.', 'What a  *Wonderful* World!' }
>>> ft_labels_1 = { '__label__' + label for label in labels }
>>> norm_labels = { 'hello_world', 'what_a_wonderful_world' }
>>> ft_labels_2 = { '__label__' + label for label in norm_labels }
>>> mlt = MultiLabelText(text)
>>> for lab in ft_labels_1:
...     mlt.add_label(lab)
>>> mlt.text == text
True
>>> mlt.labels == ft_labels_1
True
>>> fft = FromFastText()
>>> norm_mlt = fft.format(mlt)
>>> norm_mlt.text == text
True
>>> norm_mlt.labels == labels
True
>>> tft = ToFastText()
>>> to_mlt = tft.format(norm_mlt)
>>> to_mlt.text == text
True
>>> to_mlt.labels == ft_labels_2
True
>>>
>>> nft = Normalizer()
>>> n_mlt = nft.format(norm_mlt)
>>> n_mlt.text == text
True
>>> n_mlt.labels == norm_labels
True
>>> fft = FromFastText(cache_labels=True)
>>> norm_mlt = fft.format(mlt)
>>> norm_mlt.text == text
True
>>> norm_mlt.labels == labels
True
>>> tft = ToFastText(cache_labels=True)
>>> to_mlt = tft.format(norm_mlt)
>>> to_mlt.text == text
True
>>> to_mlt.labels == ft_labels_2
True
>>>
>>> nft = Normalizer(cache_labels=True)
>>> n_mlt = nft.format(norm_mlt)
>>> n_mlt.text == text
True
>>> n_mlt.labels == norm_labels
True

"""


import re
from common.prepro import normalize_label, normalize


class MultiLabelText:
    def __init__(self, text=''):
        self.text = text
        self.labels = set()

    def set_text(self, text):
        self.text = text

    def add_word(self, word):
        self.text += (' ' + word)

    def add_label(self, label):
        self.labels.add(label)


class Formatter:
    def __init__(self, cache_labels=True):
        self.cache_labels = cache_labels
        if cache_labels:
            self.cached_labels = {}

    @staticmethod
    def format_label(label):
        return label

    @staticmethod
    def format_text(text):
        return text

    def format(self, mlt):
        for_mlt = MultiLabelText(self.format_text(mlt.text))
        for lab in mlt.labels:
            if self.cache_labels:
                can_lab = self.cached_labels.get(lab)
                if not can_lab:
                    can_lab = self.format_label(lab)
                self.cached_labels[lab] = can_lab
            else:
                can_lab = self.format_label(lab)
            for_mlt.add_label(can_lab)
        return for_mlt


class FromFastText(Formatter):
    def __init__(self, cache_labels=True):
        super(self.__class__, self).__init__(cache_labels)

    @staticmethod
    def format_label(label):
        return re.sub(r'^__label__', '', label)


class ToFastText(Formatter):
    def __init__(self, cache_labels=True):
        super(self.__class__, self).__init__(cache_labels)

    @staticmethod
    def format_label(label):
        return '__label__' + normalize_label(label)

    @staticmethod
    def format_text(text):
        return normalize(text)


class Normalizer(Formatter):
    def __init__(self, cache_labels=True):
        super(self.__class__, self).__init__(cache_labels)

    @staticmethod
    def format_label(label):
        return normalize_label(label)

    @staticmethod
    def format_text(text):
        return normalize(text)



from setuptools import setup

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='yamconv',
    version='0.1.3',
    description='yamconv converts the file formats of machine learning datasets',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    keywords='machine learning',
    url='https://github.com/yam-ai/yamconv',
    author='Thomas Lee',
    author_email='thomaslee@yam.ai',
    pymodule='yamconv',
    include_page_data=True,
    zip_safe=True,
    scripts=['yamconv.py'],
    packages=['mlt', 'common'],
)



#!/usr/bin/env python3
# coding=utf-8
# Copyright 2019 YAM AI Machinery Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import getopt
import logging
from json import loads
from mlt.io import FastText2SQLite, SQLite2FastText
from common.ex import YamconvError

NUM_LINES = 1000
CACHE_LABELS = True

MLT_FASTTEXT_TO_SQLITE = 'mlt.fasttext2sqlite'
MLT_SQLITE_TO_FASTTEXT = 'mlt.sqlite2fasttext'


def main(argv):
    progname = argv[0]
    log_level = logging.WARN
    infile, outfile, convert, settings = None, None, None, None
    try:
        opts, _ = getopt.getopt(argv[1:], 'i:o:c:s:v')
        for opt, arg in opts:
            if opt == '-i':
                infile = arg
                continue
            if opt == '-o':
                outfile = arg
                continue
            if opt == '-c':
                convert = arg
                continue
            if opt == '-s':
                try:
                    settings = loads(arg)
                except Exception as e:
                    raise Exception("-s settings not in JSON")
            if opt == '-v':
                log_level = logging.INFO
                continue
    except Exception as e:
        err(progname, e)
    if not infile:
        err(progname, Exception('-i is missing'))
    if not outfile:
        err(progname, Exception('-o is missing'))
    if not convert:
        err(progname, Exception('-c is missing'))
    logger = get_logger(log_level)
    try:
        converter = get_converter(
            convert, infile, outfile,
            settings, logger, NUM_LINES)
    except Exception as e:
        err(progname, e)
    if not converter:
        err(progname,
            Exception('Unknown converter name {}'.format(convert)))
    try:
        converter.convert()
    except YamconvError as e:
        print('Failed to convert data: {}'.format(e), file=sys.stderr)


def get_converter(name, infile, outfile, settings, logger, nlines):
    converter = None
    if settings is not None:
        cache_labels = settings.get('cache_labels')
        if cache_labels not in [True, False]:
            raise YamconvError('cache_labels must be true or false')
        if cache_labels is not None:
            logger.info('cache_labels = {}'.format(cache_labels))
    if name == MLT_FASTTEXT_TO_SQLITE:
        converter = FastText2SQLite(
            infile, outfile, cache_labels=CACHE_LABELS,
            logger=logger, nlines=nlines)
    elif name == MLT_SQLITE_TO_FASTTEXT:
        converter = SQLite2FastText(
            infile, outfile, cache_labels=CACHE_LABELS,
            logger=logger, nlines=nlines)
    return converter


def get_logger(log_level):
    ch = logging.StreamHandler()
    ch.setFormatter(
        logging.Formatter(
            '[%(asctime)s|%(levelname)s] %(message)s',
            '%Y-%m-%d %H:%M:%S %z')
    )
    logger = logging.getLogger(__name__)
    logger.addHandler(ch)
    logger.setLevel(log_level)
    return logger


def err(progname, e=None):
    converter_names = [MLT_FASTTEXT_TO_SQLITE, MLT_SQLITE_TO_FASTTEXT]
    print('Usage: {} -c converter -i input_file -o ouput_file -s settings -v'.format(progname),
          file=sys.stderr)
    print('-c: converter name', file=sys.stderr)
    print('-i: input file path', file=sys.stderr)
    print('-o: output file path', file=sys.stderr)
    print('-s: converter settings in JSON', file=sys.stderr)
    print('-v: verbose', file=sys.stderr)
    print('Supported converters: {}'.format(
        ', '.join(converter_names)), file=sys.stderr)
    if e:
        print('Error: {}'.format(e), file=sys.stderr)
    sys.exit(1)


if __name__ == '__main__':
    main(sys.argv)
